@@ -2,6 +2,5 @@
 
 from rdf.formats import NTriples, NQuads 
 from rdf.graph import Statement
-from rdf.terms import Entity, Literal, BNode, IRIRef, Resource
-
-
+from rdf.terms import BNode, IRIRef, Literal
+from rdf.namespaces import OWL, RDF, RDFS, SKOS, XSD
